extern crate ctrlc;
use shakmaty::fen::Fen;
use shakmaty::{Chess, Color, Outcome, Position, Setup};
use std::io;
use std::io::{BufRead, BufReader};
use std::os::unix::net::UnixStream;
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::{Arc, Mutex};
use std::thread;
use std::time;

enum RecvPositionError {
    UserCanceledError,
    CommunicationError,
    ParsePositionError,
}

fn main() {
    let running = setupctrlc();
    let username = std::env::args().nth(1).expect("no name given");
    let public_key = std::env::args().nth(2).expect("no public key given");
    println!("Name: {}, Public key: {}", username, public_key);
    //send username and public key to server
    let mut stream = match UnixStream::connect("/tmp/clichess.socket") {
        Ok(sock) => sock,
        Err(_) => {
            println!("clichess daemon is not running.");
            return;
        }
    };
    clichess::write_to_stream(&mut stream, username).unwrap();
    clichess::write_to_stream(&mut stream, public_key).unwrap();
    let (client, chess) = get_connection_info_from_stream(&stream);
    //First prompt when connecting to the server
    println!(
        "Hello, {} !\n\r You're playing with the {} pieces",
        client.username,
        client.role.to_string()
    );
    //then we get the initial role of the connected client.
    let mut current_position = chess;
    let input_buffer = setup_input_buffer();
    //start a thread to listen to server messages
    let server_message = setup_server_message_recv(&stream).unwrap();
    loop {
        println!(
            "{}",
            clichess::board_representation(&current_position, clichess::to_color(client.side))
        );
        //check if game is over.
        if current_position.is_game_over() {
            break;
        }
        if clichess::is_player_turn(&client, current_position.turn()) {
            //it's the user turn, taking user input
            let input = read_user_input(running.clone(), input_buffer.clone());
            clichess::write_to_stream(&mut stream, String::from(input.trim())).unwrap();
            if input.trim() == "exit" {
                break;
            }
        }
        //update position after playing.
        match get_current_position(running.clone(), server_message.clone()) {
            Ok(position) => current_position = position,
            Err(RecvPositionError::UserCanceledError) => break,
            Err(RecvPositionError::CommunicationError) => break,
            Err(RecvPositionError::ParsePositionError) => break,
        };
    }
    match current_position.outcome() {
        None => println!("Bye"),
        Some(Outcome::Draw) => println!("Draw game."),
        Some(Outcome::Decisive { winner }) => {
            if winner == Color::White {
                println!("White has won the game.")
            } else {
                println!("Black has won the game.")
            }
        }
    }
}

fn setup_input_buffer() -> Arc<Mutex<String>> {
    let buf = Arc::new(Mutex::new(String::new()));
    let buf2 = buf.clone();

    thread::spawn(move || {
        loop {
            let mut buffer = String::new();
            //wait for user input
            io::stdin().read_line(&mut buffer).unwrap();
            {
                let mut user_input = buf2.lock().unwrap();
                if user_input.is_empty() {
                    *user_input = buffer;
                } else {
                    println!("It's not your turn !");
                }
            }
        }
    });
    buf
}

fn setupctrlc() -> Arc<AtomicBool> {
    let running = Arc::new(AtomicBool::new(true));
    let r = running.clone();
    ctrlc::set_handler(move || {
        r.store(false, Ordering::SeqCst);
    })
    .expect("Error setting Ctrl-C handler");
    running
}

fn setup_server_message_recv(stream: &UnixStream) -> io::Result<Arc<Mutex<String>>> {
    let buf = Arc::new(Mutex::new(String::new()));
    let buf2 = buf.clone();
    let thread_stream = stream.try_clone()?;

    thread::spawn(move || {
        loop {
            //wait for server message
            let buffer =
                clichess::read_line_from_stream(&thread_stream).expect("Error message from server");
            {
                let mut server_message = buf2.lock().unwrap();
                if server_message.is_empty() {
                    *server_message = buffer;
                } else {
                    println!("Warning: server tried to send a message while the current one has not been computed yet.");
                }
            }
        }
    });
    Ok(buf)
}

fn read_user_input(running: Arc<AtomicBool>, input_buffer: Arc<Mutex<String>>) -> String {
    //clear input before waiting for a new one
    {
        let mut user_input = input_buffer.lock().unwrap();
        user_input.clear();
    }
    let mut input = String::new();
    while running.load(Ordering::SeqCst) {
        thread::sleep(time::Duration::from_millis(10));
        {
            let user_input = input_buffer.lock().unwrap();
            if user_input.is_empty() {
                continue;
            } else {
                input = user_input.clone();
                break;
            }
        }
    }
    if running.load(Ordering::SeqCst) {
        input
    } else {
        String::from("exit")
    }
}

//wait for next position from server, then return the current board.
fn get_current_position(
    running: Arc<AtomicBool>,
    server_message: Arc<Mutex<String>>,
) -> Result<Chess, RecvPositionError> {
    let mut response = String::new();
    while running.load(Ordering::SeqCst) {
        thread::sleep(time::Duration::from_millis(10));
        {
            let mut server_response = server_message.lock().unwrap();
            if server_response.is_empty() {
                continue;
            } else {
                response = server_response.clone();
                server_response.clear();
                break;
            }
        }
    }
    if response.is_empty() {
        Err(RecvPositionError::UserCanceledError)
    } else {
        Ok(parse_position(&response))
    }
}

fn parse_position(string: &str) -> Chess {
    let setup: Fen = string.trim().parse().expect("Invalid message from server.");
    let position: Chess = setup.position().expect("Invalid message from server.");
    position
}

fn get_connection_info_from_stream(stream: &UnixStream) -> (clichess::Client, Chess) {
    println!("Read lines from stream...");
    let mut buf = String::new();
    let mut reader = BufReader::new(stream);
    //first, client as json.
    reader
        .read_line(&mut buf)
        .expect("Server closed connection.");
    let client = serde_json::from_str(&(buf.trim())).unwrap();
    buf.clear();
    reader
        .read_line(&mut buf)
        .expect("Server closed connection.");
    let chess = parse_position(buf.trim());
    (client, chess)
}
