use clichess;
use clichess::Client;
use clichess::UserRole;
use shakmaty::{fen, Chess, Color, Setup};
use std::collections::HashMap;
use std::fs;
use std::io::{BufRead, BufReader};
use std::os::unix::net::{UnixListener, UnixStream};
use std::sync::{Arc, Condvar, Mutex};

use std::thread;

#[derive(Clone)]
struct Server {
    id: usize,
    chess_position: Arc<Mutex<Chess>>,
    clients: Arc<Mutex<HashMap<usize, (Client, Arc<(Mutex<String>, Condvar)>)>>>,
    receiving_buffer: Arc<(Mutex<String>, Condvar)>,
}

fn main() {
    let chess = Arc::new(Mutex::new(Chess::default()));
    let clients = Arc::new(Mutex::new(HashMap::new()));
    let mut counter = 0;
    fs::remove_file("/tmp/clichess.socket");

    let listener = UnixListener::bind("/tmp/clichess.socket").unwrap();

    // accept connections and process them, spawning a new thread for each one
    for stream in listener.incoming() {
        match stream {
            Ok(stream) => {
                let server = Server {
                    id: counter,
                    chess_position: chess.clone(),
                    clients: clients.clone(),
                    receiving_buffer: Arc::new((Mutex::new(String::new()), Condvar::new())),
                };
                /* connection succeeded */
                thread::spawn(|| handle_client(stream, server));
                counter += 1;
            }
            Err(err) => {
                println!("Could not create a connection: {}", err);
                break;
            }
        }
    }
}

fn handle_client(mut stream: UnixStream, server: Server) {
    //create client
    let client = create_client(&server, &stream);
    //send its information to the client
    println!(
        "server {}, send first information to the client...",
        server.id
    );
    clichess::write_to_stream(&mut stream, serde_json::to_string(&client).unwrap());
    println!("server {}, information to the client sent", server.id);
    let mut player_turn: Color;
    //send current position to the client
    println!(
        "server {}, send current position to the client...",
        server.id
    );
    {
        let chess = server.chess_position.lock().unwrap();
        player_turn = chess.turn();
        clichess::write_to_stream(&mut stream, fen::fen(&*chess)).unwrap();
    }
    println!("server {}, current position to the client sent", server.id);
    loop {
        if clichess::is_player_turn(&client, player_turn) {
            //let go of the lock while waiting for user input.
            println!("server {}, waiting for client move..", server.id);
            let input;
            match clichess::read_line_from_stream(&stream) {
                Ok(i) => input = i,
                Err(e) => {
                    println!("Error while getting user input: {}", e);
                    break;
                }
            };
            if input == "exit" {
                break;
            }
            {
                let mut chess = server.chess_position.lock().unwrap();
                let clients = server.clients.lock().unwrap();
                println!("User tried to play {}", input);
                match clichess::try_to_play_move(&chess, input) {
                    Ok(played_chess) => *chess = played_chess,
                    Err(e) => println!("Error: {}", e),
                };
                let chessfen = fen::fen(&*chess);
                for (id, (_, cvar_buffer)) in clients.iter() {
                    if server.id != *id {
                        cvar_buffer.0.lock().unwrap().push_str(&chessfen);
                        cvar_buffer.1.notify_one();
                    }
                }
                clichess::write_to_stream(&mut stream, chessfen).unwrap();
                player_turn = chess.turn();
            }
        } else {
            {
                let (buffer_mutex, cvar) = &*server.receiving_buffer;
                let mut buffer = buffer_mutex.lock().unwrap();
                println!("server id: {}, waiting for move to send...", server.id);
                while buffer.is_empty() {
                    buffer = cvar.wait(buffer).unwrap();
                }
                println!("server id: {}, sending {}", server.id, buffer);
                match clichess::write_to_stream(&mut stream, buffer.clone()) {
                    Ok(()) => buffer.clear(),
                    Err(e) => {
                        println!("{}", e);
                        break;
                    }
                }
                let chess = server.chess_position.lock().unwrap();
                player_turn = chess.turn();
            }
        }
    }
    println!("Client disconnected.")
}

fn create_client(server: &Server, stream: &UnixStream) -> Client {
    println!("Creating client {}...", server.id);
    //get client name and pubkey
    println!("Read lines from stream...");
    let mut buf = String::new();
    let mut reader = BufReader::new(stream);
    //first, username.
    reader
        .read_line(&mut buf)
        .expect("Client closed connection.");
    let username = String::from(buf.trim());
    buf.clear();
    reader
        .read_line(&mut buf)
        .expect("Client closed connection.");
    let public_key = String::from(buf.trim());
    let role = match server.id {
        0 => UserRole::White,
        1 => UserRole::Black,
        _ => UserRole::Spectator,
    };
    let mut clients = server.clients.lock().unwrap();
    let client = Client {
        role,
        side: clichess::from_color(clichess::get_default_side(role)),
        username,
        public_key,
    };
    clients.insert(server.id, (client.clone(), server.receiving_buffer.clone()));
    println!("Created client {}", server.id);
    client
}
